"""Functions to help locate Eve Online log files"""
from datetime import datetime
from os import listdir
from os.path import isfile, join
from pathlib import Path
import pandas as pd

def _join_paths(filename: str, directory: str):
    return join(directory, filename)

def _date_string(date: datetime):
    return date.date().strftime('%Y-%m-%d')

def _map_log_file(filename: str):
    [name, _] = filename.split('.')
    parts =  name.split('_', 3)

    if len(parts) == 3:
        (date, time, character_id) = parts
    elif len(parts) == 2:
        (date, time, character_id) = parts + [None]

    date_time = f'{date} {time}'
    created_at = datetime.strptime(date_time, '%Y%m%d %H%M%S')

    return {
        'filename': filename,
        'character_id': character_id,
        'created_at': created_at
    }

def _log_files(directory: str):
    files = [f for f in listdir(directory) if isfile(join(directory, f))]
    files = map(_map_log_file, files)

    return pd.DataFrame(files)

LOG_HEADER_SIG = '------------------------------------------------------------'
LOG_HEADER_TITLE = 'Gamelog'

def _read_log_character_name(log_file_path):
    """Read the character name from log header"""
    with open(log_file_path, 'r', encoding='UTF8') as log_file:
        sig = log_file.readline().strip()
        if LOG_HEADER_SIG not in sig:
            return None

        title = log_file.readline().strip()
        if LOG_HEADER_TITLE not in title:
            return None

        character_line = log_file.readline().strip()
        character_name = character_line[10:]

    return character_name

def get_all_log_dates(directory: str) -> list:
    """
    Get a list of all log dates.

    :param directory: which directory to search for log files

    :returns: list of unique log file dates

    """

    result = _log_files(directory)

    dates = result['created_at'].apply(_date_string)
    dates = sorted(dates.unique(), reverse=True)

    return dates

def get_latest_log_files(directory: str) -> list:
    """
    Get all latest log files.

    :param directory: which directory to search for log files

    :returns: list of log files dict[path, filename, character_id, character, created_at]

    """
    result = _log_files(directory)

    result['path'] = result['filename'].apply(_join_paths, args=(directory, ))

    result = result.sort_values(by='created_at', ascending=False)
    result = result.groupby('character_id', dropna=True).first()
    result['character'] = result['path'].apply(_read_log_character_name)

    return result.to_dict('records')

def default_log_directory():
    """
    Get the default Eve Online logs directory.

    :returns: absolute path to logs directory

    """
    user_directory = join(Path.home(), 'Documents\\EVE\\logs\\Gamelogs')

    return user_directory



"""Squads interface"""
import gradio as gr

from smarterbombing import configuration
from smarterbombing.app.squad import Squad
from smarterbombing.webui.html_generate_statistics import generate_compound_statistics_html

def _update_aggregators(app_squad: Squad):
    app_squad.read_and_aggregate_data()

    compound_site_statistics = app_squad.get_compound_site_statistics()
    html_statistics = generate_compound_statistics_html(compound_site_statistics)

    return [
        gr.update(value=app_squad.get_outgoing_hostile_damage_graph()),
        gr.update(value=app_squad.get_incoming_hostile_damage_graph()),
        gr.update(value=html_statistics),
    ]

def _reset_input_value():
    return gr.update(value='')

def _reinitialize_readers(app_squad: Squad):
    app_squad.initialize_readers()
    return app_squad.get_open_files()

def _render_squad_graphs():
    with gr.Row(variant='compact'):
        dps_out_h = gr.LinePlot(
                x_title='Time',
                x='timestamp',
                y_title='DPS',
                y='damage',
                color='character',
                title='Outgoing Damage',
                interactive=False,
                width=530)

        dps_in_h = gr.LinePlot(
            x_title='Time',
            x='timestamp',
            y_title='DPS',
            y='damage',
            color='character',
            title='Incoming Damage',
            interactive=False,
            width=530)

    with gr.Row(variant='compact'):
        with gr.Column(variant='compact'):
            site_compound = gr.HTML()
            gr.Markdown(
                        "💡 _The average data will populate after starting at least two sites._"
                    )

    return [dps_out_h, dps_in_h, site_compound]

def _render_squad_configuration(config: dict, app_squad: Squad, squad: dict):
    def _squad_member_list(squad):
        if len(squad['characters']) == 0:
            return '⚠ No characters'

        characters = squad['characters']
        characters = map(lambda c: f'<span class="character_label">{c}</span>', characters)

        return f'<div class="character_container">{" ".join(characters)}</div>'

    def _add_character_to_squad(character):
        if character in squad['characters']:
            squad['characters'].remove(character)
        else:
            squad['characters'].append(character)

        configuration.save(config)

        return _squad_member_list(squad)

    with gr.Accordion(label='Squad Configuration', open=len(squad['characters']) == 0):
        with gr.Row(variant='compact'):
            with gr.Column():
                character_input = gr.Textbox(label='➕ Add/Remove Character')
                gr.Markdown(
                    "💡 _Type a character name and press enter to add or remove the character._"
                )
            with gr.Column(scale=2):
                character_list = gr.HTML(_squad_member_list(squad))

        open_log_files = gr.Dataframe(label='Open Logs', value=app_squad.get_open_files())
        reload_logs = gr.Button(value='Reload Logs', variant='secondary')

        reload_logs.click(lambda: _reinitialize_readers(app_squad), outputs=open_log_files)

        character_input.submit(_add_character_to_squad,
                                inputs=character_input,
                                outputs=character_list,
        ).then(_reset_input_value,
               outputs=character_input
        ).then(lambda: _reinitialize_readers(app_squad), outputs=open_log_files)

def render_squad(sb_ui: gr.Blocks, config: dict, squad: dict):
    """
    Render squad UI

    :param sb_ui: user interface
    :param config: configuration
    :param squad: squad

    """

    squad_name = squad['squad_name']

    app_squad = Squad(squad_name, squad['characters'], config)
    app_squad.initialize_aggregators()
    app_squad.initialize_readers()

    app_squad.read_and_aggregate_data()

    with gr.Tab(squad_name):
        _render_squad_configuration(config, app_squad, squad)
        graphs = _render_squad_graphs()

    sb_ui.load(lambda: _update_aggregators(app_squad), every=1, outputs=graphs)



"""Configuration helpers"""
import os
from os.path import isfile
from yaml import (
    load as ymload,
    dump as ymdump,
    Loader, Dumper)
from smarterbombing.logs.log_locator import default_log_directory

CONFIGURATION_FILE_NAME = 'smarterbombing.yaml'
CONFIGURATION_MAX_SQUADS = 10

def exists(name_prefix='') -> bool:
    """
    Check if configuration file exists.

    :returns: True if configuration exists otherwise False.

    """
    return isfile(f'{name_prefix}{CONFIGURATION_FILE_NAME}')

def create_default() -> dict:
    """
    Create configuration file with default options.

    :returns: dict with the created configuration.
    
    """

    squads = []
    for i in range(CONFIGURATION_MAX_SQUADS):
        squads.append({ 'squad_name': f'Squad {i + 1}', 'characters': [] })

    configuration = {}
    configuration['log_directory'] = default_log_directory()
    configuration['squads'] = squads
    configuration['dps_average_seconds'] = 10
    configuration['live_graph_minutes'] = 5
    configuration['site_boundary_minimum_seconds'] = 20
    configuration['ignore_list'] = ['Heavy Missile', 'Light Missile']

    return configuration


def save(configuration: dict, name_prefix=''):
    """
    Save configuration to configuration file.

    :param configuration: dict with configuration options

    """

    with open(f'{name_prefix}{CONFIGURATION_FILE_NAME}', 'w', encoding='UTF8') as config_file:
        ymdump(configuration, config_file, Dumper=Dumper)

def load(create_if_missing=False, name_prefix='') -> dict:
    """
    Load configuration from configuration file.

    :param create_if_missing: create default configuration if there is no configuration file.

    :returns: dict with configuration options or None if file doesn't exist
    
    """
    path = f'{name_prefix}{CONFIGURATION_FILE_NAME}'

    if not isfile(path):
        if not create_if_missing:
            return None

        configuration = create_default()
        save(configuration, name_prefix=name_prefix)

        return configuration

    with open(path, 'r', encoding='UTF8') as config_file:
        return ymload(config_file, Loader=Loader)

def delete(name_prefix=''):
    """
    Delete configuration file.

    :param name_prefix: append name or path in front of configuration file name

    """
    print(f'deleting: {name_prefix}{CONFIGURATION_FILE_NAME}')

    if exists(name_prefix=name_prefix):
        os.unlink(f'{name_prefix}{CONFIGURATION_FILE_NAME}')

def get_squads(configuration: dict) -> list:
    """
    Get list of squads from configuration.

    :param configuration: configuration

    :returns: list[dict] squads

    """

    return configuration.get('squads', {})

def get_log_directory(configuration: dict) -> str:
    """
    Get log directory from configuration.

    :param configuration: configuration

    :returns: str log directory
    """

    return configuration.get('log_directory', {})

def get_dps_average_seconds(configuration: dict) -> int:
    """
    Get DPS average seconds from configuration.

    :param configuration: configuration

    :returns: int dps average rolling window in seconds

    """

    return configuration.get('dps_average_seconds', 10)

def create_squad(configuration: dict, squad_name: str):
    """
    Create a new squad.

    :param configuration: configuration
    :param squad_name: the squad name

    """

    squad = {
        'squad_name': squad_name,
        'characters': []
    }

    squads = configuration.get('squads', [])
    squads.append(squad)



import pandas as pd

def generate_compound_statistics_html(compound: pd.DataFrame):
    """
    Generate a HTML table displaying compound site statistics.

    :param compound: compound site statistics

    :returns: a html string

    """

    efficiency = compound['time_efficiency']
    efficiency_percent = int(efficiency * 100)

    html = f"""
    <h4 class="scs">Average Site Efficiency</h4>
    <meter class="scs" max="1" low="0.5" optimum="1.0" high="0.7" value="{efficiency}">Efficiency</meter>
    <table class="scs">
        <tr>
            <th scope="row">Efficiency</th>
            <td>{efficiency_percent}%</td>
        </tr>
        <tr>
            <th scope="row">Total Time</th>
            <td>{compound['total_time']}</td>
        </tr>
        <tr>
            <th scope="row">Downtime</th>
            <td>{compound['total_downtime']}</td>
        </tr>
        <tr>
            <th scope="row">Effective Time</th>
            <td>{compound['total_effective_time']}</td>
        </tr>
        <tr>
            <th scope="row">Average Time</th>
            <td>{compound['average_time']}</td>
        </tr>
        <tr>
            <th scope="row">Average Downtime</th>
            <td>{compound['average_downtime']}</td>
        </tr>
        <tr>
            <th scope="row">Damage Per Second</th>
            <td>{compound['damage_per_second']}</td>
        </tr>
    </table>"""

    return html



"""Utilities to help open and read log files"""

from smarterbombing.parsing.combat_log_parser import parse_combat_log_line

def _filter_subjects(event: dict, filter_subjects: list):
    return event['subject'] not in filter_subjects

def _filter_none(event: dict):
    return event is not None

def _apply_character_info(event: dict, character: str, friendly_characters: list):
    event['character'] = character
    event['friendly_fire'] = event['subject'] in friendly_characters
    return event

class LogReader:
    """
    LogReader

    """

    def __init__(self, log_file, friendly_characters: list, filter_subjects: list):
        self.log_file = log_file
        self.friendly_characters = friendly_characters
        self.filter_subjects = filter_subjects
        self.file = None

    def is_open(self):
        """
        Check if file is open

        :returns: True if file is open, otherwise False

        """

        return self.file is not None

    def open(self, seek_to_end: bool = False):
        """
        Open the file for reading

        :param seek_to_end: Seek to end of file after opening it

        """

        self.file = open(self.log_file['path'], 'r', encoding='UTF8')

        if seek_to_end:
            self.file.seek(0, 2)

    def close(self):
        """
        Close the file

        """

        self.file.close()
        self.file = None

    def read_log_events(self):
        """
        Read and parse all available lines from file

        :returns: list of parsed events

        """

        character = self.log_file['character']

        events = filter(_filter_none, map(parse_combat_log_line, self.file.readlines()))
        events = filter(
            lambda evt: _filter_subjects(evt, self.filter_subjects), events
        )
        events = map(
            lambda evt: _apply_character_info(evt, character, self.friendly_characters), events
        )

        return events



"""User interface"""
import gradio as gr

from smarterbombing import configuration
from smarterbombing.webui.ui_squads import render_squads

def run_webui(port: int = 42069):
    """
    Run web interface

    :param port: which port to listen to

    """
    config = configuration.load(create_if_missing=True)

    custom_css = """
    .character_container {
        padding: 8px;
        line-height: 32px;
    }
    .character_label {
        background-color: #6e00ff;
        border-radius: 6px;
        padding: 4px 8px 4px 8px;
        margin: 2px;
    }

    h4.scs {
        margin-top: 8px !important;
    }

    table.scs {
        table-layout: fixed;
        width: 100%;
        border-collapse: collapse;
    }

    table.scs tr {
        border-bottom: 1px dotted black;
    }

    table.scs th {
        text-align: left !important;
    }

    table.scs td {
        text-align: right !important;
    }

    meter.scs {
        width: 100%;
        height: 32px;
    }
    """

    with gr.Blocks(title="Smarterbombing", css=custom_css) as sb_ui:
        render_squads(sb_ui, config)

        sb_ui.queue()
        sb_ui.launch(show_api=False, server_port=port)
